#######################################################################################################
##       Simulated data: Example of QL-ODE-P-splines (Frasso, Jaeger and Lambert, 2015)              ##
## This is an illustrative example: dx = (theta_1 - 2 * theta2 * t) * x^2, x(0) = 1                  ##   
#######################################################################################################

rm(list = ls()); graphics.off(); cat("\014")
#---------------------------------- Sources and libraries --------------------------------------------#

library(splines)
library(Matrix)
library(spam)

source("basis_array.R")
source("functions_to_optimize.R")
source("penalty_function.R")

#--------------------------------------- explicit solution -------------------------------------------#

f_fun = function(theta, x0, t)
{
 f = 1 / (theta[2] * t ^ 2 - theta[1] * t + (1 / x0))
 return(f)
}

#------------------------------------- min and max time ----------------------------------------------#

N_obs = 50
min_t = -0
max_t = 2

#---------------------------------- value of the parameters ------------------------------------------#

tau_sim = 500
theta_sim = c(1, 1)
x0_sim = 1

#------------------------------------- data simulation ------------------------------------------------#

set.seed(1)
t_obs = seq(min_t, max_t, length = N_obs)
x_obs = f_fun(theta_sim, x0_sim, t_obs)
y_obs = rnorm(N_obs, x_obs, 1 / sqrt(tau_sim))

#-------------------------------- B-splines for penalty matrices -------------------------------------#

length_t = 1e3
seq_t = seq(min_t, max_t, length = length_t)
ndt = 50

B_seq_t = basis_array(t = seq_t, tl = min_t, tr = max_t, ndt = ndt, deg = 3, max_derivs = 1, sparse = TRUE)
for (j in 1:length(B_seq_t))
{
  B_seq_t[[j]] = as.spam(as.matrix(B_seq_t[[j]]))
}

M = dim(B_seq_t[[1]])[2]

#----------------------------------- design matrix ----------------------------------------------------#

B = basis_array(t = t_obs, tl = min_t, tr = max_t, ndt = ndt, deg = 3, max_derivs = 0, sparse = TRUE)
B_0 = as.spam(as.matrix(B[[1]]))
t_B_0 = t(B_0)

tB_B = t_B_0 %*% B_0
tB_y = t_B_0 %*% y_obs

#---------------------------- Preliminary P-spline smoothing ------------------------------------------#

D = diff(diag(M), diff = 2)
tD_D = t(D) %*% D

las = seq(-0, 8, by = 1)
cv = ed = las * 0
i = 1

for(la in las)
{
  lam = 10 ^ la
  solve_A = solve(tB_B +  lam * tD_D)
  cy = solve_A %*% tB_y
  ed[i] = sum(diag(B_0 %*% solve_A %*% t_B_0))
  rss = sum((y_obs - B_0 %*% cy) ^ 2)
  cv[i] = rss/((N_obs - ed[i]) ^ 2)
  i = i + 1
}

lam_opt = 10 ^ las[which.min(cv)]
ck = c(solve(tB_B +  lam_opt * tD_D, tB_y)) *0 + 0.9
tau_est = (N_obs - ed[which.min(cv)])/(sum((y_obs - B_0 %*% ck) ^ 2))

#---------------------------------- stoping rules -----------------------------------------------------#

tol_gamma = 1e-4
tol_ck = 1e-4
tol_theta_tau = 1e-4

#---------------------------- QL-ODE-P-splines computations -------------------------------------------#

## Initial settings
i = 1
theta_tau = c(theta_sim * 0.5, tau_sim * 0.5) # perturbed initial parameters
gamma = 1e2
ck = cbind(ck, NULL) * 0 + 0.5 # flat initial spline coefficients
crit_gamma = crit_ck = crit_theta_tau = 1

## Loop for convergence
while(((crit_ck > tol_ck) | (crit_gamma > tol_gamma) | (crit_theta_tau > tol_theta_tau)) && (i < 100))
{
  theta_tau_old = theta_tau
  
  ## Update spline coefficients
  zandpen = c_hat_function(head(theta_tau_old, n = -1),
                           tail(theta_tau_old, n = 1),
                           gamma,
                           ck[, i],
                           length_t,
                           seq_t,
                           B_seq_t,
                           regul = 0.0)
  z = zandpen[[1]]
  pen = zandpen[[2]]
  t_z = t(z)
  y_star = B_0 %*% z
  ck = cbind(ck, z)
  crit_ck = max(abs((ck[, i] - z) / ck[, i]))
  
  ## Update gamma and tau
  PEN = as.numeric(t_z %*% pen $ R %*% z + 2 * t_z %*% pen $ r + pen $ l)
  ed = sum(diag(B_0 %*% solve((gamma / tail(theta_tau_old, n = 1)) * pen $ R +  tB_B) %*% t_B_0))
  theta_tau[length(theta_tau)] = (N_obs - ed) / sum((y_obs - y_star) ^ 2)
  den = c(PEN / ed)
  gamma_new = 1 / (den + 1e-15)
  crit_gamma = abs((gamma - gamma_new) / gamma)
  gamma = gamma_new

  ## Update ODE parameters and approx. std.err (via FIM)
  optim_parm = optim(par = head(theta_tau_old, n = -1),
                     fn = H_fun,
                     tau = tail(theta_tau_old, n = 1),
                     gamma = gamma,
                     ck = z,
                     length_t = length_t,
                     seq_t = seq_t,
                     B_seq_t = B_seq_t,
                     regul = 0.0,
                     method = "BFGS",
                     control = list(fnscale = -1, trace = FALSE),
                     hessian = TRUE)
  theta_tau[-length(theta_tau)] = optim_parm $ par
  FIM = solve(-optim_parm $ hessian)
  std.err = sqrt(diag(FIM))
  crit_theta_tau = max(abs(theta_tau - theta_tau_old) / theta_tau_old)
  
  ## Itaration and estimates 
  cat('iteration:', i,'\n')
  cat('\t', 'crit_gamma:', round(crit_gamma, 8), '\n')
  cat('\t', 'crit_ck:', round(crit_ck, 8), '\n')
  cat('\t', 'crit_theta_tau:', round(crit_theta_tau, 8), '\n')
  cat('\t', 'gamma:', round(gamma, 8), '\n')
  cat('\t', 'tau:', round(tail(theta_tau, n = 1), 8), '\n')
  cat('\t', 'theta:', head(theta_tau, n = -1), '\n')

  i = i+1
}

#-------------------------- Approx. std. err. bounds and plot the fit ---------------------------------#

## Approximate standard errors
y_star = B_0 %*% ck[, i]
rss = sum((y_obs - y_star) ^ 2)
var_hat = 1/theta_tau[length(theta_tau)]
sigma = sqrt(var_hat)
A = solve((gamma / theta_tau[length(theta_tau)]) * pen $ R + tB_B) 
Cov_ck = sigma^2 * A
Cov_y_star = B_0 %*% Cov_ck %*% t(B_0)
se = sqrt(diag(Cov_y_star))
lev = 0.95
alp = qnorm(1-(1-lev)/2)
y_star_low = y_star - alp * se
y_star_up = y_star + alp * se

##  plot results

par(mfrow = c(1,1), cex = 0.8, mar = c(3, 3, 2,2),  mgp = c(2, 0.8, 0))
plot(y_obs ~ t_obs, col = 8, main = "Example: first Order ODE",
     xlab = 't', ylab = "x(t)", pch = 16)
lines(t_obs, y_star, lwd = 2)
lines(t_obs, y_star_low, lwd = 1, lty = 2)
lines(t_obs, y_star_up, lwd = 1, lty = 2)
lines(t_obs, x_obs, lwd = 2, col = 2, lty = 2)

legend('bottomleft', legend = c(as.expression(bquote(hat(theta)[1] == .(round(theta_tau[1], 4)))),                           
                                 as.expression(bquote(hat(theta)[2] == .(round(theta_tau[2], 4)))),
                                 as.expression(bquote(gamma[] == .(gamma ))),
                                 as.expression(bquote(sigma[] == .(round(sqrt(1/theta_tau[3]), 4 ))))),
       box.lty = 0, cex = 0.85)

legend('topright', legend =  c("Observations", "Estimated dynamics", "Analytic solution"),
       box.lty = 0, cex = 1, pch = c(16, -1, -1, -1),
       lty = c(0, 1, 2), col = c(8, 1, 2),
       lwd = c(0, 2, 2))



