#######################################################################################################
## Estimation SA-AV dynamics using coupled van der Pol systems (dos Santos et al., 2004)             ##
## Realistic data obtained by adding a Gaussian noise to the solution of the ODE system              ##
## Warning: computationally slow since we estimate 7 unknown ODE parameters                          ##
## Estimates based on QL-ODE-P-splines (Frasso, Jaeger and Lambert, 2015)                            ##
#######################################################################################################

#---------------------------------------- ODE model --------------------------------------------------#
#ddot(x1) = kappa * (x1 - w1) * (x1 - w2) * dot(x1) - b1 * x1 + c1 * (x2 - x1)
#ddot(x2) = kappa * (x2 - w1) * (x2 - w2) * dot(x2) - b1 * x2 + c1 * (x1 - x2)

#----------------------------------------- library ---------------------------------------------------#

rm(list = ls()); graphics.off(); cat("\014")

library(splines)
library(Matrix)
library(spam)
library(deSolve)
library(orthogonalsplinebasis)

#--------------------------------------------- source files ------------------------------------------#
source("basis_array.R")
source("functions_to_optimize.R")
source("penalty_function.R")

#--------------------------------------- min and max time ---------------------------------------------#

min_t = 0
max_t = 20

#----------------------------------------------- value of the parameters ------------------------------#

set.seed(1)
tau_sim = 30
b1 = 1.5
b2 = 0.1
kappa = -1.8
w1 = -0.2
w2 = 2
c1 = 0.05
c2 = 0.55

#--------------------------------------- numerical solution -------------------------------------------#

pars = c(b1, b2, kappa, w1, w2, c1, c2)

coupled_VdP_model = function(Time, State, Pars)
{
  with(as.list(c(State, Pars, t)),
       {
         dx1 = x2
         dx2 = Pars[3] * (x1 - Pars[4]) * (x1 - Pars[5]) * x2 - Pars[1] * x1 + Pars[6] * (x3 - x1)
         dx3 = x4
         dx4 = Pars[3] * (x3 - Pars[4]) * (x3 - Pars[5]) * x4 - Pars[2] * x3 + Pars[7] * (x1 - x3)
         
         return(list(c(dx1, dx2, dx3, dx4)))  
       })
}

ini  = c(x1 = 1, x2 = 1, x3 = 1, x4 = 1)
N_obs = 200
t_obs = seq(min_t, max_t, len = N_obs)
out = ode(ini, t_obs, coupled_VdP_model, pars)


#----------------------------------------- data simuation ---------------------------------------------#

y_obs_1 = rnorm(N_obs, out[, "x2"], sd = 1/sqrt(tau_sim))
y_obs_2 = rnorm(N_obs, out[, "x4"], sd = 1/sqrt(tau_sim))
y_obs = c(y_obs_1, y_obs_2)
N_obs = length(y_obs_1) * 2

#---------------------------------------- elementary penalty matrices ---------------------------------#

length_t = 1e3
seq_t = seq(min_t, max_t, length = length_t)
ndt = 250

B_seq_t = list()
B_seq_t[[2]] = as.spam(as.matrix(basis_array(x = seq_t, xl = min_t, xr = max_t, ndx = ndt, deg = 4,
                                             max_derivs = 2, sparse = TRUE)[[1]])) 
B_seq_t[[1]] = as.spam(as.matrix(integrate_basis(seq_t, xl = min_t, xr = max_t, ndx = ndt, deg = 4,
                                                 B2 = B_seq_t[[2]])))
B_seq_t[[3]] = as.spam(as.matrix(basis_array(x = seq_t, xl = min_t, xr = max_t, ndx = ndt, deg = 4,
                                             max_derivs = 2, sparse = TRUE)[[2]]))

#--------------------------------------- stoping rules ------------------------------------------------#

tol_gamma = 5e-4
tol_ck = 5e-4
tol_theta_tau = 5e-4

#-------------------------------------- design matrix and diff pen ------------------------------------#

B = basis_array(x = t_obs, xl = min_t, xr = max_t, ndx = ndt, deg = 4, max_derivs = 0, sparse = TRUE)

B_0 = (as.matrix(B[[1]]))
t_B_0 = t(B_0)

B_i = as.spam(as.matrix(integrate_basis(t_obs, xl = min_t, xr = max_t, ndx = ndt, deg = 4, B2 = B_0))) 


tB_B = t_B_0 %*% B_0

tB_y = t_B_0 %*% cbind(y_obs_1, y_obs_2)

M = dim(tB_B)[2]

D = diff(diag(M), diff = 2)
tD_D = t(D) %*% D

#----------------------------------------- Initial smoothing raw data ---------------------------------#

# smoothing for x1 and x3
lam_opt = 1e-1
ck = c(solve(tB_B + lam_opt * tD_D, tB_y, tol = 1e-25)) 

#--------------------------------- Quasi ODE-P-spline -------------------------------------------------#

i = 1
ck = cbind(ck, NULL)

theta_tau = c(pars * 0.85, tau_sim) # "perturbed" initial parameters

crit_gamma = crit_ck = crit_theta_tau = 1
gamma = 1e5

while(((crit_ck > tol_ck) | (crit_gamma > tol_gamma) | (crit_theta_tau > tol_theta_tau)) && (i < 100))
{
  flush.console()
  cat('iteration:', i,'\n')
  
  theta_tau_old = theta_tau
  
  zandpen = c_hat_function(theta = theta_tau[-length(theta_tau)],
                           tau = theta_tau[length(theta_tau)],
                           gamma,
                           ck[, i],
                           length_t,
                           seq_t,
                           B_seq_t,
                           regul = 0.0)
  
  z = zandpen[[1]]
  pen = zandpen[[2]]
  t_z = t(z)
  
  y_star = B_0 %*% matrix(z, ncol = 2)
  PEN = (t_z %*% pen $ R %*% z + 2 * t_z %*% pen $ r + pen $ l)
  tau_1 = tau_2 = theta_tau[length(theta_tau)]
  
  A = as.matrix(gamma * pen $ R +
                  rbind(cbind(tB_B   * tau_1, tB_B * 0),
                        cbind(tB_B * 0, tB_B * tau_2))
  )
  
  
  solve_A = solve(A, tol = 1e-50)
  
  ed_1 = sum(diag(B_0 %*% solve_A[1:M, 1:M] %*% ( tau_1 * t_B_0)))
  ed_2 = sum(diag(B_0 %*% solve_A[M + 1:M,  M + 1:M] %*% (tau_2 * t_B_0)))
  
  tau_1 = ((N_obs/2) - ed_1) / sum((y_obs_1 - c(y_star[, 1])) ^ 2)
  tau_2 = ((N_obs/2) - ed_2) / sum((y_obs_2 - c(y_star[, 2])) ^ 2)
  
  theta_tau[length(theta_tau)] = (N_obs - (ed_1 + ed_2) ) / sum( c((y_obs - y_star) ^ 2) )
  den = c(PEN / (ed_1 + ed_2))
  gamma_new = 1 / (den + 1e-10)
  
  crit_gamma = abs((gamma - gamma_new) / gamma)
  gamma = gamma_new
  
  ck = cbind(ck, z)
  crit_ck = max(abs((ck[, i] - z) / ck[, i]))
  
  optim_parm = optim(par = theta_tau[-length(theta_tau)],
                     fn = H_fun,
                     tau =  theta_tau[length(theta_tau)],
                     gamma = gamma,
                     ck = z,
                     length_t = length_t,
                     seq_t = seq_t,
                     B_seq_t = B_seq_t,
                     regul = 0.0,
                     control = list(fnscale = -1),
                     method = 'BFGS',                         
                     hessian = TRUE)
  
  theta_tau[-length(theta_tau)] = optim_parm $ par
  crit_theta_tau = max(abs((theta_tau - theta_tau_old) / theta_tau_old))
  
  cat('\t', 'crit_theta_tau:', round(crit_theta_tau, 8), '\n')
  cat('\t', 'crit_gamma:', round(crit_gamma, 8), '\n')
  cat('\t', 'gamma:', round(gamma, 8), '\n')
  cat('\t', 'crit_ck:', round(crit_ck, 8), '\n')
  cat('\t', 'tau:', round(tail(theta_tau, n = 1), 8), '\n')
  cat('\t', 'theta:', head(theta_tau, n = -1), '\n')
  
  i = i+1
}

#------------------------------------ Plots -----------------------------------------------------------#    
FIM = solve(-optim_parm $ hessian)
std_err = sqrt(diag(FIM))
tau_est = theta_tau[length( theta_tau)] 

A = as.matrix(gamma * pen $ R +
                rbind(cbind(tB_B * tau_est, tB_B * 0),
                      cbind(tB_B * 0, tB_B * tau_est)))
solve_A = solve(A, tol = 1e-50) * (1/tau_est)  

bounds = (B_seq_t[[1]] %*% solve_A[1:M, 1:M] %*% t(B_seq_t[[1]]) * tau_est)
lev = 0.99
alp = qnorm(1-(1-lev)/2)

low_bound = B_seq_t[[1]] %*% z[1:M] - alp * sqrt(diag(bounds))
up_bound = B_seq_t[[1]] %*% z[1:M] + alp * sqrt(diag(bounds))   

par(mfrow = c(2,2), mar = c(3.5, 3.5, 2, 1), cex.main = 1, mgp = c(2, 0.8, 0))

# x1
plot(t_obs, out[, "x1"], type = "n", xlab = "Time", ylab = expression(x[1]),
     main = "Sino-atrial node", ylim = range( out[, "x1"]) - c(0.2, -0.2)) 
lines(seq_t, B_seq_t[[1]] %*% z[1:M], col = 1, lwd = 2, lty = 1)
lines(seq_t, low_bound, col = 2, lty = 2)
lines(seq_t, up_bound, col = 2, lty = 2)

lines(t_obs, out[, "x1"], type = "l", col = 8, lty = 2, lwd = 2) 
legend("topleft", c("Estimates", "Numerical solution", "Std. Err. bands"),
       col = c("black", "gray", "red"), pch = c(-1, -1 -1),
       lty = c(1, 2, 2), lwd = c(2, 2, 1), box.lty = 0, cex = 0.75)

# x2  
bounds = B_seq_t[[2]] %*% solve_A[1:M, 1:M] %*% t(B_seq_t[[2]] * tau_est)

low_bound = B_seq_t[[2]]  %*% z[1:M] - alp * sqrt(diag(bounds))
up_bound = B_seq_t[[2]]  %*% z[1:M] + alp * sqrt(diag(bounds))            

plot(t_obs, y_obs_1, pch = 1, cex = 0.8, col = 1, main = "Sino-atrial node",
     xlab = "Time", ylab = expression(x[2]), ylim = range(y_obs) - c(0.25, -0.25))
lines(seq_t, B_seq_t[[2]]  %*% z[1:M], lwd = 2)
lines(seq_t, low_bound, col = 2, lty = 2)
lines(seq_t, up_bound, col = 2, lty = 2)

lines(t_obs, out[, "x2"], col = 8, lty = 8, lwd = 2)
legend("topleft", c("Observations", "Estimates", "Numerical solution", 
                    "Std. Err. bands"),
       col = c("black", "black", "gray", "red"), pch = c(1, -1, -1, -1),
       lty = c(0, 1, 2, 2), lwd = c(0, 2, 2, 1), box.lty = 0, cex = 0.75) 

# x3               
bounds =  B_seq_t[[1]] %*% solve_A[M + 1:M, M + 1:M] %*% t(B_seq_t[[1]] * tau_est)

low_bound = B_seq_t[[1]]  %*% z[M + 1:M] - alp * sqrt(diag(bounds))
up_bound = B_seq_t[[1]]  %*% z[M + 1:M] + alp * sqrt(diag(bounds))            

plot(t_obs, out[, "x3"], type = "n", xlab = "Time", ylab = expression(x[3]),
     main = "Atrio-ventricular node" , ylim = range( out[, "x3"]) - c(0.2, -0.2) ) 
lines(seq_t, B_seq_t[[1]]  %*% z[M + 1:M], col = 1, lwd = 2, lty = 1)
lines(seq_t, low_bound, col = 2, lty = 2)
lines(seq_t, up_bound, col = 2, lty = 2)

lines(t_obs, out[, "x3"], col = 8, lty = 2, lwd = 2) 
legend("top", c("Estimates", "Numerical solution", "Std. Err. bands"),
       col = c("black", "gray", "red"), pch = c(-1, -1),
       lty = c(1, 2, 2), lwd = c(2, 2, 1), box.lty = 0, cex = 0.75)

# x4
bounds =  B_seq_t[[2]] %*% solve_A[M + 1:M, M + 1:M] %*% t(B_seq_t[[2]] * tau_est)

low_bound = B_seq_t[[2]]  %*% z[M + 1:M] - alp * sqrt(diag(bounds))
up_bound = B_seq_t[[2]]  %*% z[M + 1:M] + alp * sqrt(diag(bounds))            


plot(t_obs, y_obs_2, pch = 1, cex = 0.8, col = 1, main = "Atrio-ventricular node",   
     xlab = "Time", ylab = expression(x[4]), ylim = range(y_obs) - c(0.25, -0.25))
lines(seq_t, B_seq_t[[2]]  %*% z[M + 1:M], lwd = 2)
lines(seq_t, low_bound, col = 2, lty = 2)
lines(seq_t, up_bound, col = 2, lty = 2)


lines(t_obs, out[, "x4"], col = 8, lty = 2, lwd = 2.8)
legend("top", c("Observations", "Estimated hare dynamics", "Numerical solution",
                "Std. Err. bands"),
       col = c("black", "black", "gray", "red"), pch = c(1, -1, -1, -1),
       lty = c(0, 1, 2, 2), lwd = c(0, 2, 2, 1), box.lty = 0, cex = 0.75)
