#######################################################################################################
##          Penalty function for QL-ODE-P-splines (Frasso, Jaeger and Lambert, 2015)                 ##
## This function returns the terms to build a quasilinearized penalty for the coupled                ##
## Van der Pol system.                                                                               ##  
#######################################################################################################

## Function arguments:
## theta = current estimates ODE parameters
## ck = current estimates spline coefficients
## length_t = length grid B-spline for penalty matrix
## seq_t = grid of time points (B-spline  for penalty)
## B_seq_t = list of B-splines terms for the construction of the penalty
## min_t, max_t = min and max values of observed times

pen_fun = function(theta, ck, length_t, seq_t, B_seq_t, min_t, max_t)
{ 

 b1 = theta[1]
 b2 = theta[2]
 kappa = theta[3]
 w1 = theta[4]
 w2 = theta[5]
 c1 = theta[6]
 c2 = theta[7]
                
 B_x_0_seq_t = as.matrix(B_seq_t[[1]])
 B_x_1_seq_t = as.matrix(B_seq_t[[2]])
 B_x_2_seq_t = as.matrix(B_seq_t[[3]])
 
 ck1 = ck[1:M]
 ck2 = ck[-c(1:M)]
 
 B_x_0_seq_t_times_ck1 = B_x_0_seq_t %*% ck1   
 B_x_1_seq_t_times_ck1 = B_x_1_seq_t %*% ck1

 B_x_0_seq_t_times_ck2 = B_x_0_seq_t %*% ck2
 B_x_1_seq_t_times_ck2 = B_x_1_seq_t %*% ck2                             
 
 # Define first ODE pen: w_row_statefun_degderiv
 
 w_1_1_2 = rep(1, length_t)
 w_1_1_1 = - as.vector(kappa * (B_x_0_seq_t_times_ck1 - w1) * (B_x_0_seq_t_times_ck1 - w2) )
 w_1_1_0 = - as.vector(kappa * (2 * B_x_0_seq_t_times_ck1 - w1 - w2) * B_x_1_seq_t_times_ck1 ) + c1 + b1
 w_1_2_0 = w_1_1_2 * 0 - c1
 
 H1 = cbind(w_1_1_2 * B_x_2_seq_t + w_1_1_1 * B_x_1_seq_t + w_1_1_0 * B_x_0_seq_t,  B_x_0_seq_t * w_1_2_0)

 h1 = as.vector((kappa * (2 * B_x_0_seq_t_times_ck1 - w1 - w2) * B_x_1_seq_t_times_ck1) * B_x_0_seq_t_times_ck1)
 
 w_2_2_2 = rep(1, length_t)
 w_2_2_1 = -as.vector(kappa * (B_x_0_seq_t_times_ck2 - w1) * (B_x_0_seq_t_times_ck2 - w2))
 w_2_2_0 = -as.vector(kappa * (2 * B_x_0_seq_t_times_ck2 - w1 - w2) * B_x_1_seq_t_times_ck2) + c2 + b2
 w_2_1_0 = w_1_1_2 * 0 - c2
 
 H2 = cbind(B_x_0_seq_t * w_2_1_0 , w_2_2_2 * B_x_2_seq_t + w_2_2_1 * B_x_1_seq_t + w_2_2_0 * B_x_0_seq_t)
 h2 = as.vector((kappa * (2 * B_x_0_seq_t_times_ck2 - w1 - w2) * B_x_1_seq_t_times_ck2) * B_x_0_seq_t_times_ck2)

 R1 = (max_t - min_t) /
        (length_t - 1) *
        (t(H1) %*% H1)

 R2 = (max_t - min_t) /
        (length_t - 1) *
        (t(H2) %*% H2)

 R = R1 + R2 

 r1 = (max_t - min_t) /
        (length_t - 1) *
        (t(H1) %*% h1)
 r2 = (max_t - min_t) /
        (length_t - 1) *
        (t(H2) %*% h2)
 r = c(r1 + r2)
 
 l1 = (max_t - min_t) /
      (length_t - 1) *
      (t(h1) %*% h1)
 l2 = (max_t - min_t) /
      (length_t - 1) *
      (t(h2) %*% h2)
 l = c(l1 + l2) 

 
 return(list(R = R, r = r, l = l ))
}

