#######################################################################################################
##       Simulated data: Example of QL-ODE-P-splines (Frasso, Jaeger and Lambert, 2015)              ##
## This is an illustrative example for the Van der Pol system                                        ##   
#######################################################################################################

rm(list = ls()); graphics.off(); cat("\014")
#------------------------------- library --------------------------------------------------------------#

library(splines)
library(Matrix)
library(spam)
library(deSolve)

#------------------------------ source files ----------------------------------------------------------#

source("basis_array.R")
source("functions_to_optimize.R")
source("penalty_function.R")

#---------------------------- value of the parameters -------------------------------------------------#

tau_sim = 50
theta_sim = 1
x0_sim = 1
x0_prime_sim = - 1
min_t = 0
max_t = 10

#---------------------------- numerical solution ------------------------------------------------------#

parameters = c(theta = theta_sim)

state = c(X = x0_sim, Y = x0_sim - 1 / 3 * (x0_sim ^ 3) - 1 / theta_sim * x0_prime_sim)

vdp = function(t, state, parameters)
{
 with(as.list(c(state, parameters)),
  {
   dX = theta * (X - 1 / 3 * X * X * X - Y)
   dY = 1 / theta * X
   return(list(c(dX, dY)))
  })
}

t = seq(min_t, max_t, by = 1e-4)
out = ode(y = state, times = t, func = vdp, parms = parameters)


#-------------------------  B-spline matrices for the penalty -----------------------------------------#

length_t = 1e3    
seq_t = seq(min_t, max_t, length = length_t)
ndt = 50

B_seq_t = basis_array(t = seq_t, tl = min_t, tr = max_t, ndt = ndt, deg = 3, max_derivs = 2, sparse = TRUE)
for (j in 1:length(B_seq_t))
{
  B_seq_t[[j]] = as.spam(as.matrix(B_seq_t[[j]]))
}

M = dim(B_seq_t[[1]])[2]


#---------------------------------- elements for initial smoothing ------------------------------------#

D = diff(diag(M), diff = 2)
tD_D = t(D) %*% D

#----------------------------------------- stoping rules ----------------------------------------------#

tol_gamma = 1e-4
tol_ck = 1e-4
tol_theta_tau = 1e-4


#--------------------------------------- Data simulation ----------------------------------------------#

set.seed(1)
N_obs = 100
t_obs = seq(min_t, max_t, length = N_obs)
l = c()

for (i in 1:N_obs)
{
  l[i] = which.min(abs(t_obs[i] - t))
}
t_obs = t[l]
x_obs = out[l, "X"]

y_obs = rnorm(N_obs, x_obs, 1 / sqrt(tau_sim))
  
#----------------------------------------- design matrix ----------------------------------------------#
  
B = basis_array(t = t_obs, tl = min_t, tr = max_t, ndt = ndt, deg = 3, max_derivs = 1, sparse = TRUE) 
B_0 = as.spam(as.matrix(B[[1]]))
t_B_0 = t(B_0)
  
tB_B = t_B_0 %*% B_0
tB_y = t_B_0 %*% y_obs
  
#-------------------------- Set initial parameters and spline coeff -----------------------------------#

gamma = 1e2
theta_tau = c(theta_sim * 0.5, tau_sim * 0.5) # perturbed parameters

# first smoothing for x1
las = seq(0, 8, by = 1)
cv = las * 0
i = 1

for(la in las)
{
  lam = 10 ^ la
  solve_A = solve(tB_B +  lam * tD_D)
  cy = solve_A %*% tB_y
  ed = sum(diag(B_0 %*% solve_A %*% t_B_0))
  rss = sum((y_obs - B_0 %*% cy) ^ 2)
  cv[i] = rss/((N_obs - ed) ^ 2)
  i = i + 1
}
    
lam_opt = 10 ^ las[which.min(cv)]
ck = c(solve(tB_B +  lam_opt * tD_D, tB_y))

# first smoothing for x2

x1_tilde = as.vector(B[[1]] %*% ck)
tBprime = t(B[[2]])
tBprime_x1_tilde = tBprime %*% x1_tilde
tBprime_Bprime = tBprime %*% B[[2]]
    
las = seq(0, 8, by = 1)
cv = las * 0
i = 1
    
for(la in las)
{
 lam = 10 ^ la
 solve_A = solve(tBprime_Bprime +  lam * tD_D)
 cy = solve_A %*% (1 / theta_tau[1] * tBprime_x1_tilde)
 ed = sum(diag(B[[2]] %*% solve_A %*% tBprime))
 rss = sum((as.vector(1 / theta_tau[1] * x1_tilde) - B[[2]] %*% as.vector(cy)) ^ 2)
 cv[i] = rss/((N_obs - ed) ^ 2)
 i = i + 1
}

lam_opt = 10 ^ las[which.min(cv)]
ck = c(ck, as.vector(solve(tBprime_Bprime +  lam_opt * tD_D, 1 / theta_tau[1] * tBprime_x1_tilde)))
ck = ck * 0 + 0.9 # flat initial spline coefficients

#------------------------------------ QL-ODE-P-spline ------------------------------------------------#

i = 1

ck = cbind(ck, NULL)
theta_tau[length(theta_tau)] = N_obs / sum((y_obs - B_0 %*% ck[1:M, ]) ^ 2)
crit_gamma = crit_ck = crit_theta_tau = 1
    
while(((crit_ck > tol_ck) | (crit_gamma > tol_gamma) | (crit_theta_tau > tol_theta_tau)) && (i < 100))
{
  theta_tau_old = theta_tau
  
  zandpen = c_hat_function(head(theta_tau, n = -1),
                           tail(theta_tau, n = 1),
                           gamma,
                           ck[, i],
                           length_t,
                           seq_t,
                           B_seq_t,
                           regul = 0.0)
  z = zandpen[[1]]
  pen = zandpen[[2]]
  t_z = t(z)
  
  y_star = B_0 %*% z[1:M]
  
  PEN = as.numeric(t_z %*% pen $ R %*% z + 2 * t_z %*% pen $ r + pen $ l)
  
  A = gamma * pen $ R +
      cbind(rbind(tail(theta_tau, n = 1) * tB_B, array(0, dim = c(M, M))),
            array(0, dim = c(2 * M, M)))
  solve_A = solve(A)
      
  ed = sum(diag(B_0 %*% solve_A[1:M, 1:M] %*% (tail(theta_tau, n = 1) * t_B_0)))
  theta_tau[length(theta_tau)] = (N_obs - ed) / sum((y_obs - y_star) ^ 2)
  
  den = c(PEN / ed)
  
  gamma_new = 1 / (den + 1e-15)
  
  crit_gamma = abs((gamma - gamma_new) / gamma)
  gamma = gamma_new
  
  ck = cbind(ck, z)
  crit_ck = max(abs((ck[, i] - z) / ck[, i]))
  
  optim_parm = optim(par = head(theta_tau, n = -1),
                     fn = H_fun,
                     tau = tail(theta_tau, n = 1),
                     gamma = gamma,
                     ck = z,
                     length_t = length_t,
                     seq_t = seq_t,
                     B_seq_t = B_seq_t,
                     regul = 0.0,
                     control = list(fnscale = -1, trace = FALSE),
                     method = "BFGS",
                     hessian = TRUE)
      
  theta_tau[-length(theta_tau)] = optim_parm $ par
  FIM = solve(-optim_parm $ hessian)
      
  crit_theta_tau = max(abs(theta_tau - theta_tau_old) / theta_tau_old)
  
  cat('iteration:', i,'\n')
  cat('\t', 'crit_gamma:', round(crit_gamma, 8), '\n')
  cat('\t', 'crit_ck:', round(crit_ck, 8), '\n')
  cat('\t', 'crit_theta_tau:', round(crit_theta_tau, 8), '\n')
  cat('\t', 'gamma:', round(gamma, 8), '\n')
  cat('\t', 'tau:', round(tail(theta_tau, n = 1), 8), '\n')
  cat('\t', 'theta:', head(theta_tau, n = -1), '\n')
  
  i = i+1
}

#-------------------------- Approx. std. err. bounds and plot the fit ---------------------------------#

## Approximate standard errors
y_star_1 = B_0 %*% ck[1:M, i]
y_star_2 = B_0 %*% ck[M + 1:M, i]
var_hat = 1/theta_tau[length(theta_tau)]
sigma = sqrt(var_hat)
A = gamma * pen $ R +
    cbind(rbind(tail(theta_tau, n = 1) * tB_B, array(0, dim = c(M, M))),
          array(0, dim = c(2 * M, M)))
solve_A = solve(A)

## y_star_1
Cov_ck = sigma^2 * solve_A[1:M, 1:M]
Cov_y_star = B_0 %*% Cov_ck %*% t(theta_tau[length(theta_tau)] * B_0)
se = sqrt(diag(Cov_y_star))
lev = 0.99
alp = qnorm(1-(1-lev)/2)
y_star_1_low = y_star_1 - alp * se
y_star_1_up = y_star_1 + alp * se

## y_star_2
Cov_ck = sigma^2 * solve_A[M + 1:M, M + 1:M]
Cov_y_star = B_0 %*% Cov_ck %*% t(theta_tau[length(theta_tau)] * B_0)
se = sqrt(diag(Cov_y_star))
lev = 0.99
alp = qnorm(1-(1-lev)/2)
y_star_2_low = y_star_2 - alp * se
y_star_2_up = y_star_2 + alp * se

##  plot results

par(mfrow = c(2,1), cex = 0.8, mar = c(3, 3, 2,2),  mgp = c(2, 0.8, 0))
plot(y_obs ~ t_obs, col = 8, main = "Example: Van der Pol system",
     xlab = 't', ylab = "x(t)", pch = 16)
lines(t_obs, y_star_1, lwd = 2)
lines(t_obs, y_star_1_low, lwd = 1, lty = 2)
lines(t_obs, y_star_1_up, lwd = 1, lty = 2)
lines(t_obs, x_obs, lwd = 2, col = 2, lty = 2)

legend('topright', legend =  c("Observations", "Estimated dynamics", "Analytic solution"),
       box.lty = 0, cex = 1, pch = c(16, -1, -1, -1),
       lty = c(0, 1, 2), col = c(8, 1, 2),
       lwd = c(0, 2, 2))

plot(t_obs, out[l, "Y"], lwd = 2, col = 2, lty = 2, type = "n")
lines(t_obs, y_star_2, lwd = 2)
lines(t_obs, y_star_2_low, lwd = 1, lty = 2)
lines(t_obs, y_star_2_up, lwd = 1, lty = 2)
lines(t_obs, out[l, "Y"], lwd = 2, col = 2, lty = 2)



