#######################################################################################################
##          Penalty function for QL-ODE-P-splines (Frasso, Jaeger and Lambert, 2015)                 ##
## This function returns the terms to build a quasilinearized penalty for the Van der Pol system.    ##
#######################################################################################################

## Function arguments:
## theta = current estimates ODE parameters
## ck = current estimates spline coefficients
## length_t = length grid B-spline for penalty matrix
## seq_t = grid of time points (B-spline  for penalty)
## B_seq_t = list of B-splines terms for the construction of the penalty
## min_t, max_t = min and max values of observed times


pen_fun = function(theta, ck, length_t, seq_t, B_seq_t, min_t, max_t)
{         
 B_x_0_seq_t = as.matrix(B_seq_t[[1]])
 B_x_1_seq_t = as.matrix(B_seq_t[[2]])

 B_x_0_seq_t_times_ck1 = B_x_0_seq_t %*% ck[1:M]

 B_x_0_seq_t_times_ck1_sq = B_x_0_seq_t_times_ck1 * B_x_0_seq_t_times_ck1
 B_x_0_seq_t_times_ck1_third = B_x_0_seq_t_times_ck1_sq * B_x_0_seq_t_times_ck1

 w111 = rep(1, length_t)
 w112 = as.vector(theta * (-1 + B_x_0_seq_t_times_ck1_sq))
 w12 = theta * w111

 w21 = - 1 / theta * w111
 w22 = w111

 H1 = cbind(w111 * B_x_1_seq_t + w112 * B_x_0_seq_t,
            w12 * B_x_0_seq_t) 
 h1 = - 2 / 3 * theta * B_x_0_seq_t_times_ck1_third

 H2 = cbind(w21 * B_x_0_seq_t,
            w22 * B_x_1_seq_t)

 h2 = rep(0, length_t)

 R1 = (max_t - min_t) /
      (length_t - 1) *
      (t(H1)%*% H1)
 R2 = (max_t - min_t) /
      (length_t - 1) *
      (t(H2)%*% H2)

 r1 = (max_t - min_t) /
      (length_t - 1) *
      (t(H1) %*% h1)
 r2 = (max_t - min_t) /
      (length_t - 1) *
      (t(H2) %*% h2)

 l1 = (max_t - min_t) /
      (length_t - 1) *
      (t(h1) %*% h1)
 l2 = (max_t - min_t) /
      (length_t - 1) *
      (t(h2) %*% h2)

 R = R1 + R2
 r = r1 + r2
 l = l1 + l2

 return(list(R = R, r = c(r), l = c(l)))
}